@@ -1,7 +1,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2018 The OpenLDAP Foundation.
+ * Copyright 1999-2024 The OpenLDAP Foundation.
  * Portions Copyright 1999 John C. Quillan.
  * Portions Copyright 2002 myinternet Limited.
  * All rights reserved.
@@ -16,7 +16,7 @@
  */
 
 #include "perl_back.h"
-#include "../config.h"
+#include "../slap-config.h"
 
 static ConfigDriver perl_cf;
 
@@ -43,6 +43,7 @@ static ConfigTable perlcfg[] = {
 		(void *)offsetof(PerlBackend, pb_filter_search_results),
 		"( OLcfgDbAt:11.3 NAME 'olcPerlFilterSearchResults' "
 			"DESC 'Filter search results before returning to client' "
+			"EQUALITY booleanMatch "
 			"SYNTAX OMsBoolean SINGLE-VALUE )", NULL, NULL },
 	{ "perlModuleConfig", "args", 2, 0, 0,
 		ARG_MAGIC|PERL_CONFIG, perl_cf, 
@@ -171,6 +172,7 @@ perl_cf(
 			break;
 		}
 	} else {
+		PERL_SET_CONTEXT( PERL_INTERPRETER );
 		switch( c->type ) {
 		case PERL_MODULE:
 			snprintf( eval_str, EVAL_BUF_SIZE, "use %s;", c->argv[1] );
@@ -181,7 +183,7 @@ perl_cf(
 
 				snprintf( c->cr_msg, sizeof( c->cr_msg ), "%s: error %s",
 					c->log, SvPV(ERRSV, len ));
-				Debug( LDAP_DEBUG_ANY, "%s\n", c->cr_msg, 0, 0 );
+				Debug( LDAP_DEBUG_ANY, "%s\n", c->cr_msg );
 				rc = 1;
 			} else {
 				dSP; ENTER; SAVETMPS;
