@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2005-2022 The OpenLDAP Foundation.
+ * Copyright 2005-2024 The OpenLDAP Foundation.
  * Portions copyright 2004-2005 Symas Corporation.
  * All rights reserved.
  *
@@ -1567,7 +1567,7 @@ accesslog_response(Operation *op, SlapReply *rs)
 	lo = logops+logop+EN_OFFSET;
 
 	/* can't do anything if logDB isn't open */
-	if ( !SLAP_DBOPEN( li->li_db ) ) {
+	if ( !li->li_db || !SLAP_DBOPEN( li->li_db ) ) {
 		goto skip;
 	}
 
@@ -2497,7 +2497,13 @@ accesslog_db_root(
 				attr_merge_one( e, slap_schema.si_ad_entryCSN,
 					&a->a_vals[0], &a->a_nvals[0] );
 				attr_merge( e, a->a_desc, a->a_vals, a->a_nvals );
+
+				/* Populate minCSN */
 				attr_merge( e, ad_minCSN, a->a_vals, a->a_nvals );
+				ber_bvarray_dup_x( &li->li_mincsn, a->a_vals, NULL );
+				li->li_numcsns = a->a_numvals;
+				li->li_sids = slap_parse_csn_sids( li->li_mincsn, li->li_numcsns, NULL );
+				slap_sort_csn_sids( li->li_mincsn, li->li_sids, li->li_numcsns, NULL );
 			}
 			be_entry_release_rw( op, e_ctx, 0 );
 		}
